package org.eclipse.jdt.internal.compiler.lookup;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.Arrays;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeParameter;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.ImplicitNullAnnotationVerifier;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodVerifier;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeVariableBinding;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;
import org.eclipse.jdt.internal.compiler.util.SimpleSet;
import org.eclipse.jdt.internal.compiler.util.Sorting;

class MethodVerifier15 extends MethodVerifier {
  MethodVerifier15(LookupEnvironment environment) {
    super(environment);
  }
  
  protected boolean canOverridingMethodDifferInErasure(MethodBinding overridingMethod, MethodBinding inheritedMethod) {
    if (overridingMethod.areParameterErasuresEqual(inheritedMethod))
      return false; 
    if (overridingMethod.declaringClass.isRawType())
      return false; 
    return true;
  }
  
  boolean canSkipInheritedMethods() {
    if (this.type.superclass() != null && (
      this.type.superclass().isAbstract() || this.type.superclass().isParameterizedType()))
      return false; 
    return (this.type.superInterfaces() == Binding.NO_SUPERINTERFACES);
  }
  
  boolean canSkipInheritedMethods(MethodBinding one, MethodBinding two) {
    return !(two != null && (
      !TypeBinding.equalsEquals((TypeBinding)one.declaringClass, (TypeBinding)two.declaringClass) || one.declaringClass.isParameterizedType()));
  }
  
  void checkConcreteInheritedMethod(MethodBinding concreteMethod, MethodBinding[] abstractMethods) {
    super.checkConcreteInheritedMethod(concreteMethod, abstractMethods);
    boolean analyseNullAnnotations = this.environment.globalOptions.isAnnotationBasedNullAnalysisEnabled;
    AbstractMethodDeclaration srcMethod = null;
    if (analyseNullAnnotations && this.type.equals(concreteMethod.declaringClass))
      srcMethod = concreteMethod.sourceMethod(); 
    boolean useTypeAnnotations = this.environment.usesNullTypeAnnotations();
    boolean hasReturnNonNullDefault = (analyseNullAnnotations && concreteMethod.hasNonNullDefaultFor(16, useTypeAnnotations));
    boolean hasParameterNonNullDefault = (analyseNullAnnotations && concreteMethod.hasNonNullDefaultFor(8, useTypeAnnotations));
    for (int i = 0, l = abstractMethods.length; i < l; i++) {
      MethodBinding abstractMethod = abstractMethods[i];
      if (concreteMethod.isVarargs() != abstractMethod.isVarargs())
        problemReporter().varargsConflict(concreteMethod, abstractMethod, this.type); 
      MethodBinding originalInherited = abstractMethod.original();
      if (TypeBinding.notEquals(originalInherited.returnType, concreteMethod.returnType) && 
        !isAcceptableReturnTypeOverride(concreteMethod, abstractMethod))
        problemReporter().unsafeReturnTypeOverride(concreteMethod, originalInherited, this.type); 
      if (originalInherited.declaringClass.isInterface() && ((
        TypeBinding.equalsEquals((TypeBinding)concreteMethod.declaringClass, (TypeBinding)this.type.superclass) && this.type.superclass.isParameterizedType() && !areMethodsCompatible(concreteMethod, originalInherited)) || 
        this.type.superclass.erasure().findSuperTypeOriginatingFrom((TypeBinding)originalInherited.declaringClass) == null))
        this.type.addSyntheticBridgeMethod(originalInherited, concreteMethod.original()); 
      if (analyseNullAnnotations && !concreteMethod.isStatic() && !abstractMethod.isStatic())
        checkNullSpecInheritance(concreteMethod, srcMethod, hasReturnNonNullDefault, hasParameterNonNullDefault, true, abstractMethod, abstractMethods, (Scope)this.type.scope, null); 
    } 
  }
  
  void checkForBridgeMethod(MethodBinding currentMethod, MethodBinding inheritedMethod, MethodBinding[] allInheritedMethods) {
    if (currentMethod.isVarargs() != inheritedMethod.isVarargs())
      problemReporter(currentMethod).varargsConflict(currentMethod, inheritedMethod, this.type); 
    MethodBinding originalInherited = inheritedMethod.original();
    if (TypeBinding.notEquals(originalInherited.returnType, currentMethod.returnType) && 
      !isAcceptableReturnTypeOverride(currentMethod, inheritedMethod))
      problemReporter(currentMethod).unsafeReturnTypeOverride(currentMethod, originalInherited, this.type); 
    SyntheticMethodBinding syntheticMethodBinding = this.type.addSyntheticBridgeMethod(originalInherited, currentMethod.original());
    if (syntheticMethodBinding != null) {
      for (int i = 0, l = (allInheritedMethods == null) ? 0 : allInheritedMethods.length; i < l; i++) {
        if (allInheritedMethods[i] != null && detectInheritedNameClash(originalInherited, allInheritedMethods[i].original()))
          return; 
      } 
      MethodBinding[] current = (MethodBinding[])this.currentMethods.get(((MethodBinding)syntheticMethodBinding).selector);
      for (int j = current.length - 1; j >= 0; j--) {
        MethodBinding thisMethod = current[j];
        if (thisMethod.areParameterErasuresEqual((MethodBinding)syntheticMethodBinding) && TypeBinding.equalsEquals(thisMethod.returnType.erasure(), ((MethodBinding)syntheticMethodBinding).returnType.erasure())) {
          problemReporter(thisMethod).methodNameClash(thisMethod, inheritedMethod.declaringClass.isRawType() ? inheritedMethod : inheritedMethod.original(), 1);
          return;
        } 
      } 
    } 
  }
  
  void checkForNameClash(MethodBinding currentMethod, MethodBinding inheritedMethod) {
    if (inheritedMethod.isStatic() || currentMethod.isStatic()) {
      MethodBinding original = inheritedMethod.original();
      if ((this.type.scope.compilerOptions()).complianceLevel >= 3342336L && currentMethod.areParameterErasuresEqual(original))
        problemReporter(currentMethod).methodNameClashHidden(currentMethod, inheritedMethod.declaringClass.isRawType() ? inheritedMethod : original); 
      return;
    } 
    if (!detectNameClash(currentMethod, inheritedMethod, false)) {
      TypeBinding[] currentParams = currentMethod.parameters;
      TypeBinding[] inheritedParams = inheritedMethod.parameters;
      int length = currentParams.length;
      if (length != inheritedParams.length)
        return; 
      for (int i = 0; i < length; i++) {
        if (TypeBinding.notEquals(currentParams[i], inheritedParams[i]) && (
          currentParams[i].isBaseType() != inheritedParams[i].isBaseType() || !inheritedParams[i].isCompatibleWith(currentParams[i])))
          return; 
      } 
      ReferenceBinding[] interfacesToVisit = null;
      int nextPosition = 0;
      ReferenceBinding superType = inheritedMethod.declaringClass;
      ReferenceBinding[] itsInterfaces = superType.superInterfaces();
      if (itsInterfaces != Binding.NO_SUPERINTERFACES) {
        nextPosition = itsInterfaces.length;
        interfacesToVisit = itsInterfaces;
      } 
      superType = superType.superclass();
      while (superType != null && superType.isValidBinding()) {
        MethodBinding[] methods = superType.getMethods(currentMethod.selector);
        for (int m = 0, n = methods.length; m < n; m++) {
          MethodBinding substitute = computeSubstituteMethod(methods[m], currentMethod);
          if (substitute != null && !isSubstituteParameterSubsignature(currentMethod, substitute) && detectNameClash(currentMethod, substitute, true))
            return; 
        } 
        if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES)
          if (interfacesToVisit == null) {
            interfacesToVisit = itsInterfaces;
            nextPosition = interfacesToVisit.length;
          } else {
            int itsLength = itsInterfaces.length;
            if (nextPosition + itsLength >= interfacesToVisit.length)
              System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
            for (int a = 0; a < itsLength; a++) {
              ReferenceBinding next = itsInterfaces[a];
              int b = 0;
              while (true) {
                if (b >= nextPosition) {
                  interfacesToVisit[nextPosition++] = next;
                  break;
                } 
                if (TypeBinding.equalsEquals((TypeBinding)next, (TypeBinding)interfacesToVisit[b]))
                  break; 
                b++;
              } 
            } 
          }  
        superType = superType.superclass();
      } 
      for (int j = 0; j < nextPosition; j++) {
        superType = interfacesToVisit[j];
        if (superType.isValidBinding()) {
          MethodBinding[] methods = superType.getMethods(currentMethod.selector);
          for (int m = 0, n = methods.length; m < n; m++) {
            MethodBinding substitute = computeSubstituteMethod(methods[m], currentMethod);
            if (substitute != null && !isSubstituteParameterSubsignature(currentMethod, substitute) && detectNameClash(currentMethod, substitute, true))
              return; 
          } 
          if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
            int itsLength = itsInterfaces.length;
            if (nextPosition + itsLength >= interfacesToVisit.length)
              System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
            for (int a = 0; a < itsLength; a++) {
              ReferenceBinding next = itsInterfaces[a];
              int b = 0;
              while (true) {
                if (b >= nextPosition) {
                  interfacesToVisit[nextPosition++] = next;
                  break;
                } 
                if (TypeBinding.equalsEquals((TypeBinding)next, (TypeBinding)interfacesToVisit[b]))
                  break; 
                b++;
              } 
            } 
          } 
        } 
      } 
    } 
  }
  
  void checkInheritedMethods(MethodBinding inheritedMethod, MethodBinding otherInheritedMethod) {
    if (inheritedMethod.isStatic())
      return; 
    if (this.environment.globalOptions.complianceLevel < 3342336L && inheritedMethod.declaringClass.isInterface())
      return; 
    detectInheritedNameClash(inheritedMethod.original(), otherInheritedMethod.original());
  }
  
  void checkInheritedMethods(MethodBinding[] methods, int length, boolean[] isOverridden, boolean[] isInherited) {
    boolean continueInvestigation = true;
    MethodBinding concreteMethod = null;
    MethodBinding abstractSuperClassMethod = null;
    boolean playingTrump = false;
    int i;
    for (i = 0; i < length; i++) {
      if (!(methods[i]).declaringClass.isInterface() && 
        TypeBinding.notEquals((TypeBinding)(methods[i]).declaringClass, (TypeBinding)this.type) && 
        methods[i].isAbstract()) {
        abstractSuperClassMethod = methods[i];
        break;
      } 
    } 
    for (i = 0; i < length; i++) {
      if (isInherited[i] && !methods[i].isAbstract()) {
        if (methods[i].isDefaultMethod() && 
          abstractSuperClassMethod != null && 
          areParametersEqual(abstractSuperClassMethod, methods[i]) && 
          concreteMethod == null) {
          playingTrump = true;
          continue;
        } 
        playingTrump = false;
        if (concreteMethod != null) {
          if (isOverridden[i] && areMethodsCompatible(concreteMethod, methods[i]))
            continue; 
          if (TypeBinding.equalsEquals((TypeBinding)concreteMethod.declaringClass, (TypeBinding)(methods[i]).declaringClass) && 
            concreteMethod.typeVariables.length != (methods[i]).typeVariables.length) {
            if (concreteMethod.typeVariables == Binding.NO_TYPE_VARIABLES && 
              concreteMethod.original() == methods[i])
              continue; 
            if ((methods[i]).typeVariables == Binding.NO_TYPE_VARIABLES && 
              methods[i].original() == concreteMethod)
              continue; 
          } 
          problemReporter().duplicateInheritedMethods(this.type, concreteMethod, methods[i], 
              (this.environment.globalOptions.sourceLevel >= 3407872L));
          continueInvestigation = false;
        } 
        concreteMethod = methods[i];
      } 
      continue;
    } 
    if (continueInvestigation) {
      if (playingTrump) {
        if (!this.type.isAbstract()) {
          problemReporter().abstractMethodMustBeImplemented(this.type, abstractSuperClassMethod);
          return;
        } 
      } else if (concreteMethod != null && concreteMethod.isDefaultMethod() && 
        this.environment.globalOptions.complianceLevel >= 3407872L && 
        !checkInheritedDefaultMethods(methods, isOverridden, length)) {
        return;
      } 
      super.checkInheritedMethods(methods, length, isOverridden, isInherited);
    } 
  }
  
  boolean checkInheritedDefaultMethods(MethodBinding[] methods, boolean[] isOverridden, int length) {
    if (length < 2)
      return true; 
    boolean ok = true;
    for (int i = 0; i < length; i++) {
      if (methods[i].isDefaultMethod() && !isOverridden[i])
        for (int j = 0; j < length; j++) {
          if (j != i && !isOverridden[j] && 
            isMethodSubsignature(methods[i], methods[j]) && 
            !doesMethodOverride(methods[i], methods[j]) && !doesMethodOverride(methods[j], methods[i])) {
            problemReporter().inheritedDefaultMethodConflictsWithOtherInherited(this.type, methods[i], methods[j]);
            ok = false;
            break;
          } 
        }  
    } 
    return ok;
  }
  
  boolean checkInheritedReturnTypes(MethodBinding method, MethodBinding otherMethod) {
    if (areReturnTypesCompatible(method, otherMethod))
      return true; 
    if (isUnsafeReturnTypeOverride(method, otherMethod)) {
      if (!method.declaringClass.implementsInterface(otherMethod.declaringClass, false))
        problemReporter(method).unsafeReturnTypeOverride(method, otherMethod, this.type); 
      return true;
    } 
    return false;
  }
  
  void checkAgainstInheritedMethods(MethodBinding currentMethod, MethodBinding[] methods, int length, MethodBinding[] allInheritedMethods) {
    super.checkAgainstInheritedMethods(currentMethod, methods, length, allInheritedMethods);
    CompilerOptions options = this.environment.globalOptions;
    if (options.isAnnotationBasedNullAnalysisEnabled && (
      currentMethod.tagBits & 0x1000L) == 0L) {
      AbstractMethodDeclaration srcMethod = null;
      if (this.type.equals(currentMethod.declaringClass))
        srcMethod = currentMethod.sourceMethod(); 
      boolean useTypeAnnotations = this.environment.usesNullTypeAnnotations();
      boolean hasReturnNonNullDefault = currentMethod.hasNonNullDefaultFor(16, useTypeAnnotations);
      boolean hasParameterNonNullDefault = currentMethod.hasNonNullDefaultFor(8, useTypeAnnotations);
      for (int i = length; --i >= 0;) {
        if (!currentMethod.isStatic() && !methods[i].isStatic())
          checkNullSpecInheritance(currentMethod, srcMethod, hasReturnNonNullDefault, hasParameterNonNullDefault, true, methods[i], methods, (Scope)this.type.scope, null); 
      } 
    } 
  }
  
  void checkNullSpecInheritance(MethodBinding currentMethod, AbstractMethodDeclaration srcMethod, boolean hasReturnNonNullDefault, boolean hasParameterNonNullDefault, boolean complain, MethodBinding inheritedMethod, MethodBinding[] allInherited, Scope scope, ImplicitNullAnnotationVerifier.InheritedNonNullnessInfo[] inheritedNonNullnessInfos) {
    int i = complain & (currentMethod.isConstructor() ? 0 : 1);
    if (!hasReturnNonNullDefault && !hasParameterNonNullDefault && i == 0 && !this.environment.globalOptions.inheritNullAnnotations) {
      currentMethod.tagBits |= 0x1000L;
      return;
    } 
    if (TypeBinding.notEquals((TypeBinding)currentMethod.declaringClass, (TypeBinding)this.type) && (
      currentMethod.tagBits & 0x1000L) == 0L)
      this.buddyImplicitNullAnnotationsVerifier.checkImplicitNullAnnotations(currentMethod, srcMethod, i, scope); 
    super.checkNullSpecInheritance(currentMethod, srcMethod, hasReturnNonNullDefault, hasParameterNonNullDefault, i, inheritedMethod, allInherited, scope, inheritedNonNullnessInfos);
  }
  
  void reportRawReferences() {
    CompilerOptions compilerOptions = this.type.scope.compilerOptions();
    if (compilerOptions.sourceLevel < 3211264L || 
      compilerOptions.reportUnavoidableGenericTypeProblems)
      return; 
    Object[] methodArray = this.currentMethods.valueTable;
    for (int s = methodArray.length; --s >= 0; ) {
      if (methodArray[s] == null)
        continue; 
      MethodBinding[] current = (MethodBinding[])methodArray[s];
      for (int i = 0, length = current.length; i < length; i++) {
        MethodBinding currentMethod = current[i];
        if ((currentMethod.modifiers & 0x30000000) == 0) {
          AbstractMethodDeclaration methodDecl = currentMethod.sourceMethod();
          if (methodDecl == null)
            return; 
          TypeBinding[] parameterTypes = currentMethod.parameters;
          Argument[] arguments = methodDecl.arguments;
          for (int j = 0, size = currentMethod.parameters.length; j < size; j++) {
            TypeBinding parameterType = parameterTypes[j];
            Argument arg = arguments[j];
            if (parameterType.leafComponentType().isRawType() && 
              compilerOptions.getSeverity(536936448) != 256 && (
              arg.type.bits & 0x40000000) == 0)
              methodDecl.scope.problemReporter().rawTypeReference((ASTNode)arg.type, parameterType); 
          } 
          if (!methodDecl.isConstructor() && methodDecl instanceof MethodDeclaration) {
            TypeReference returnType = ((MethodDeclaration)methodDecl).returnType;
            TypeBinding methodType = currentMethod.returnType;
            if (returnType != null && 
              methodType.leafComponentType().isRawType() && 
              compilerOptions.getSeverity(536936448) != 256 && (
              returnType.bits & 0x40000000) == 0)
              methodDecl.scope.problemReporter().rawTypeReference((ASTNode)returnType, methodType); 
          } 
        } 
      } 
    } 
  }
  
  public void reportRawReferences(MethodBinding currentMethod, MethodBinding inheritedMethod) {
    CompilerOptions compilerOptions = this.type.scope.compilerOptions();
    if (compilerOptions.sourceLevel < 3211264L || 
      compilerOptions.reportUnavoidableGenericTypeProblems)
      return; 
    AbstractMethodDeclaration methodDecl = currentMethod.sourceMethod();
    if (methodDecl == null)
      return; 
    TypeBinding[] parameterTypes = currentMethod.parameters;
    TypeBinding[] inheritedParameterTypes = inheritedMethod.parameters;
    Argument[] arguments = methodDecl.arguments;
    for (int j = 0, size = currentMethod.parameters.length; j < size; j++) {
      TypeBinding parameterType = parameterTypes[j];
      TypeBinding inheritedParameterType = inheritedParameterTypes[j];
      Argument arg = arguments[j];
      if (parameterType.leafComponentType().isRawType())
        if (inheritedParameterType.leafComponentType().isRawType()) {
          arg.binding.tagBits |= 0x200L;
        } else if (compilerOptions.getSeverity(536936448) != 256 && (
          arg.type.bits & 0x40000000) == 0) {
          methodDecl.scope.problemReporter().rawTypeReference((ASTNode)arg.type, parameterType);
        }  
    } 
    TypeReference returnType = null;
    if (!methodDecl.isConstructor() && methodDecl instanceof MethodDeclaration && (returnType = ((MethodDeclaration)methodDecl).returnType) != null) {
      TypeBinding inheritedMethodType = inheritedMethod.returnType;
      TypeBinding methodType = currentMethod.returnType;
      if (methodType.leafComponentType().isRawType() && 
        !inheritedMethodType.leafComponentType().isRawType())
        if ((returnType.bits & 0x40000000) == 0 && 
          compilerOptions.getSeverity(536936448) != 256)
          methodDecl.scope.problemReporter().rawTypeReference((ASTNode)returnType, methodType);  
    } 
  }
  
  void checkMethods() {
    boolean mustImplementAbstractMethods = mustImplementAbstractMethods();
    boolean skipInheritedMethods = (mustImplementAbstractMethods && canSkipInheritedMethods());
    boolean isOrEnclosedByPrivateType = this.type.isOrEnclosedByPrivateType();
    char[][] methodSelectors = this.inheritedMethods.keyTable;
    for (int s = methodSelectors.length; --s >= 0; ) {
      if (methodSelectors[s] == null)
        continue; 
      MethodBinding[] current = (MethodBinding[])this.currentMethods.get(methodSelectors[s]);
      MethodBinding[] inherited = (MethodBinding[])this.inheritedMethods.valueTable[s];
      inherited = Sorting.concreteFirst(inherited, inherited.length);
      if (current == null && !isOrEnclosedByPrivateType) {
        int length = inherited.length;
        for (int j = 0; j < length; j++)
          (inherited[j].original()).modifiers |= 0x8000000; 
      } 
      if (current == null && this.type.isPublic()) {
        int length = inherited.length;
        for (int j = 0; j < length; j++) {
          MethodBinding inheritedMethod = inherited[j];
          if (inheritedMethod.isPublic() && !inheritedMethod.declaringClass.isInterface() && !inheritedMethod.declaringClass.isPublic())
            this.type.addSyntheticBridgeMethod(inheritedMethod.original()); 
        } 
      } 
      if (current == null && skipInheritedMethods)
        continue; 
      if (inherited.length == 1 && current == null) {
        if (mustImplementAbstractMethods && inherited[0].isAbstract())
          checkAbstractMethod(inherited[0]); 
        continue;
      } 
      int index = -1;
      int inheritedLength = inherited.length;
      MethodBinding[] matchingInherited = new MethodBinding[inheritedLength];
      MethodBinding[] foundMatch = new MethodBinding[inheritedLength];
      boolean[] skip = new boolean[inheritedLength];
      boolean[] isOverridden = new boolean[inheritedLength];
      boolean[] isInherited = new boolean[inheritedLength];
      Arrays.fill(isInherited, true);
      if (current != null)
        for (int j = 0, length1 = current.length; j < length1; j++) {
          MethodBinding currentMethod = current[j];
          MethodBinding[] nonMatchingInherited = null;
          for (int k = 0; k < inheritedLength; k++) {
            MethodBinding inheritedMethod = computeSubstituteMethod(inherited[k], currentMethod);
            if (inheritedMethod != null)
              if (foundMatch[k] == null && isSubstituteParameterSubsignature(currentMethod, inheritedMethod)) {
                skip[k] = couldMethodOverride(currentMethod, inheritedMethod);
                isOverridden[k] = couldMethodOverride(currentMethod, inheritedMethod);
                matchingInherited[++index] = inheritedMethod;
                foundMatch[k] = currentMethod;
              } else {
                checkForNameClash(currentMethod, inheritedMethod);
                if (inheritedLength > 1) {
                  if (nonMatchingInherited == null)
                    nonMatchingInherited = new MethodBinding[inheritedLength]; 
                  nonMatchingInherited[k] = inheritedMethod;
                } 
              }  
          } 
          if (index >= 0) {
            checkAgainstInheritedMethods(currentMethod, matchingInherited, index + 1, nonMatchingInherited);
            for (; index >= 0; matchingInherited[index--] = null);
          } 
        }  
      int i;
      for (i = 0; i < inheritedLength; i++) {
        MethodBinding matchMethod = foundMatch[i];
        if (matchMethod == null && current != null && this.type.isPublic()) {
          MethodBinding methodBinding = inherited[i];
          if (methodBinding.isPublic() && !methodBinding.declaringClass.isInterface() && !methodBinding.declaringClass.isPublic())
            this.type.addSyntheticBridgeMethod(methodBinding.original()); 
        } 
        if (!isOrEnclosedByPrivateType && matchMethod == null && current != null)
          (inherited[i].original()).modifiers |= 0x8000000; 
        MethodBinding inheritedMethod = inherited[i];
        for (int j = i + 1; j < inheritedLength; j++) {
          MethodBinding otherInheritedMethod = inherited[j];
          if (matchMethod != foundMatch[j] || matchMethod == null)
            if (!canSkipInheritedMethods(inheritedMethod, otherInheritedMethod))
              if (TypeBinding.notEquals((TypeBinding)inheritedMethod.declaringClass, (TypeBinding)otherInheritedMethod.declaringClass))
                if (!isSkippableOrOverridden(inheritedMethod, otherInheritedMethod, skip, isOverridden, isInherited, j))
                  if (isSkippableOrOverridden(otherInheritedMethod, inheritedMethod, skip, isOverridden, isInherited, i));    
        } 
      } 
      for (i = 0; i < inheritedLength; i++) {
        MethodBinding matchMethod = foundMatch[i];
        if (!skip[i]) {
          MethodBinding inheritedMethod = inherited[i];
          if (matchMethod == null)
            matchingInherited[++index] = inheritedMethod; 
          for (int j = i + 1; j < inheritedLength; j++) {
            if (foundMatch[j] == null) {
              MethodBinding otherInheritedMethod = inherited[j];
              if (matchMethod != foundMatch[j] || matchMethod == null)
                if (!canSkipInheritedMethods(inheritedMethod, otherInheritedMethod)) {
                  MethodBinding replaceMatch;
                  if ((replaceMatch = findReplacedMethod(inheritedMethod, otherInheritedMethod)) != null) {
                    matchingInherited[++index] = replaceMatch;
                    skip[j] = true;
                  } else if ((replaceMatch = findReplacedMethod(otherInheritedMethod, inheritedMethod)) != null) {
                    matchingInherited[++index] = replaceMatch;
                    skip[j] = true;
                  } else if (matchMethod == null) {
                    checkInheritedMethods(inheritedMethod, otherInheritedMethod);
                  } 
                }  
            } 
          } 
          if (index != -1) {
            if (index > 0) {
              boolean[] matchingIsOverridden, matchingIsInherited;
              int length = index + 1;
              if (length != inheritedLength) {
                matchingIsOverridden = new boolean[length];
                matchingIsInherited = new boolean[length];
                for (int k = 0; k < length; k++) {
                  for (int m = 0; m < inheritedLength; m++) {
                    if (matchingInherited[k] == inherited[m]) {
                      matchingIsOverridden[k] = isOverridden[m];
                      matchingIsInherited[k] = isInherited[m];
                      break;
                    } 
                  } 
                } 
              } else {
                matchingIsOverridden = isOverridden;
                matchingIsInherited = isInherited;
              } 
              checkInheritedMethods(matchingInherited, length, matchingIsOverridden, matchingIsInherited);
            } else if (mustImplementAbstractMethods && matchingInherited[0].isAbstract() && matchMethod == null) {
              checkAbstractMethod(matchingInherited[0]);
            } 
            for (; index >= 0; matchingInherited[index--] = null);
          } 
        } 
      } 
    } 
  }
  
  boolean isSkippableOrOverridden(MethodBinding specific, MethodBinding general, boolean[] skip, boolean[] isOverridden, boolean[] isInherited, int idx) {
    boolean specificIsInterface = specific.declaringClass.isInterface();
    boolean generalIsInterface = general.declaringClass.isInterface();
    if (!specificIsInterface && generalIsInterface) {
      if (!specific.isAbstract() && isParameterSubsignature(specific, general)) {
        isInherited[idx] = false;
        return true;
      } 
      if (isInterfaceMethodImplemented(general, specific, general.declaringClass)) {
        skip[idx] = true;
        isOverridden[idx] = true;
        return true;
      } 
    } else if (specificIsInterface == generalIsInterface && 
      specific.declaringClass.isCompatibleWith((TypeBinding)general.declaringClass) && isMethodSubsignature(specific, general)) {
      skip[idx] = true;
      isOverridden[idx] = true;
      return true;
    } 
    return false;
  }
  
  MethodBinding findReplacedMethod(MethodBinding specific, MethodBinding general) {
    MethodBinding generalSubstitute = computeSubstituteMethod(general, specific);
    if (generalSubstitute != null && (
      !specific.isAbstract() || general.isAbstract() || (general.isDefaultMethod() && specific.declaringClass.isClass())) && 
      isSubstituteParameterSubsignature(specific, generalSubstitute))
      return generalSubstitute; 
    return null;
  }
  
  void checkTypeVariableMethods(TypeParameter typeParameter) {
    char[][] methodSelectors = this.inheritedMethods.keyTable;
    for (int s = methodSelectors.length; --s >= 0; ) {
      if (methodSelectors[s] == null)
        continue; 
      MethodBinding[] inherited = (MethodBinding[])this.inheritedMethods.valueTable[s];
      if (inherited.length == 1)
        continue; 
      int index = -1;
      MethodBinding[] matchingInherited = new MethodBinding[inherited.length];
      for (int i = 0, length = inherited.length; i < length; i++) {
        for (; index >= 0; matchingInherited[index--] = null);
        MethodBinding inheritedMethod = inherited[i];
        if (inheritedMethod != null) {
          matchingInherited[++index] = inheritedMethod;
          for (int j = i + 1; j < length; j++) {
            MethodBinding otherInheritedMethod = inherited[j];
            if (!canSkipInheritedMethods(inheritedMethod, otherInheritedMethod)) {
              otherInheritedMethod = computeSubstituteMethod(otherInheritedMethod, inheritedMethod);
              if (otherInheritedMethod != null && isSubstituteParameterSubsignature(inheritedMethod, otherInheritedMethod)) {
                matchingInherited[++index] = otherInheritedMethod;
                inherited[j] = null;
              } 
            } 
          } 
        } 
        if (index > 0) {
          MethodBinding first = matchingInherited[0];
          int count = index + 1;
          while (--count > 0) {
            MethodBinding match = matchingInherited[count];
            MethodBinding interfaceMethod = null, implementation = null;
            if (first.declaringClass.isInterface()) {
              interfaceMethod = first;
            } else if (first.declaringClass.isClass()) {
              implementation = first;
            } 
            if (match.declaringClass.isInterface()) {
              interfaceMethod = match;
            } else if (match.declaringClass.isClass()) {
              implementation = match;
            } 
            if (interfaceMethod != null && implementation != null && !isAsVisible(implementation, interfaceMethod))
              problemReporter().inheritedMethodReducesVisibility(typeParameter, implementation, new MethodBinding[] { interfaceMethod }); 
            if (areReturnTypesCompatible(first, match))
              continue; 
            if (first.declaringClass.isInterface() && match.declaringClass.isInterface() && areReturnTypesCompatible(match, first))
              continue; 
            break;
          } 
          if (count > 0) {
            problemReporter().inheritedMethodsHaveIncompatibleReturnTypes((ASTNode)typeParameter, matchingInherited, index + 1);
            break;
          } 
        } 
      } 
    } 
  }
  
  boolean detectInheritedNameClash(MethodBinding inherited, MethodBinding otherInherited) {
    if (!inherited.areParameterErasuresEqual(otherInherited))
      return false; 
    if (TypeBinding.notEquals(inherited.returnType.erasure(), otherInherited.returnType.erasure()))
      return false; 
    if (TypeBinding.notEquals(inherited.declaringClass.erasure(), otherInherited.declaringClass.erasure())) {
      if (inherited.declaringClass.findSuperTypeOriginatingFrom((TypeBinding)otherInherited.declaringClass) != null)
        return false; 
      if (otherInherited.declaringClass.findSuperTypeOriginatingFrom((TypeBinding)inherited.declaringClass) != null)
        return false; 
    } 
    problemReporter().inheritedMethodsHaveNameClash(this.type, inherited, otherInherited);
    return true;
  }
  
  boolean detectNameClash(MethodBinding current, MethodBinding inherited, boolean treatAsSynthetic) {
    MethodBinding methodToCheck = inherited;
    MethodBinding original = methodToCheck.original();
    if (!current.areParameterErasuresEqual(original))
      return false; 
    int severity = 1;
    if (this.environment.globalOptions.complianceLevel == 3276800L)
      if (TypeBinding.notEquals(current.returnType.erasure(), original.returnType.erasure()))
        severity = 0;  
    if (!treatAsSynthetic) {
      MethodBinding[] currentNamesakes = (MethodBinding[])this.currentMethods.get(inherited.selector);
      if (currentNamesakes.length > 1)
        for (int i = 0, length = currentNamesakes.length; i < length; i++) {
          MethodBinding currentMethod = currentNamesakes[i];
          if (currentMethod != current && doesMethodOverride(currentMethod, inherited)) {
            methodToCheck = currentMethod;
            break;
          } 
        }  
    } 
    original = methodToCheck.original();
    if (!current.areParameterErasuresEqual(original))
      return false; 
    original = inherited.original();
    problemReporter(current).methodNameClash(current, inherited.declaringClass.isRawType() ? inherited : original, severity);
    if (severity == 0)
      return false; 
    return true;
  }
  
  boolean doTypeVariablesClash(MethodBinding one, MethodBinding substituteTwo) {
    return (one.typeVariables != Binding.NO_TYPE_VARIABLES && !(substituteTwo instanceof org.eclipse.jdt.internal.compiler.lookup.ParameterizedGenericMethodBinding));
  }
  
  SimpleSet findSuperinterfaceCollisions(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
    ReferenceBinding[] interfacesToVisit = null;
    int nextPosition = 0;
    ReferenceBinding[] itsInterfaces = superInterfaces;
    if (itsInterfaces != Binding.NO_SUPERINTERFACES) {
      nextPosition = itsInterfaces.length;
      interfacesToVisit = itsInterfaces;
    } 
    boolean isInconsistent = this.type.isHierarchyInconsistent();
    ReferenceBinding superType = superclass;
    while (superType != null && superType.isValidBinding()) {
      isInconsistent |= superType.isHierarchyInconsistent();
      if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES)
        if (interfacesToVisit == null) {
          interfacesToVisit = itsInterfaces;
          nextPosition = interfacesToVisit.length;
        } else {
          int itsLength = itsInterfaces.length;
          if (nextPosition + itsLength >= interfacesToVisit.length)
            System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
          for (int a = 0; a < itsLength; a++) {
            ReferenceBinding next = itsInterfaces[a];
            int b = 0;
            while (true) {
              if (b >= nextPosition) {
                interfacesToVisit[nextPosition++] = next;
                break;
              } 
              if (TypeBinding.equalsEquals((TypeBinding)next, (TypeBinding)interfacesToVisit[b]))
                break; 
              b++;
            } 
          } 
        }  
      superType = superType.superclass();
    } 
    for (int i = 0; i < nextPosition; i++) {
      superType = interfacesToVisit[i];
      isInconsistent |= superType.isHierarchyInconsistent();
      if (superType.isValidBinding() && (itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
        int itsLength = itsInterfaces.length;
        if (nextPosition + itsLength >= interfacesToVisit.length)
          System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition); 
        for (int a = 0; a < itsLength; a++) {
          ReferenceBinding next = itsInterfaces[a];
          int b = 0;
          while (true) {
            if (b >= nextPosition) {
              interfacesToVisit[nextPosition++] = next;
              break;
            } 
            if (TypeBinding.equalsEquals((TypeBinding)next, (TypeBinding)interfacesToVisit[b]))
              break; 
            b++;
          } 
        } 
      } 
    } 
    if (!isInconsistent)
      return null; 
    SimpleSet copy = null;
    for (int j = 0; j < nextPosition; j++) {
      ReferenceBinding current = interfacesToVisit[j];
      if (current.isValidBinding()) {
        TypeBinding erasure = current.erasure();
        for (int k = j + 1; k < nextPosition; k++) {
          ReferenceBinding next = interfacesToVisit[k];
          if (next.isValidBinding() && TypeBinding.equalsEquals(next.erasure(), erasure)) {
            if (copy == null)
              copy = new SimpleSet(nextPosition); 
            copy.add(interfacesToVisit[j]);
            copy.add(interfacesToVisit[k]);
          } 
        } 
      } 
    } 
    return copy;
  }
  
  boolean isAcceptableReturnTypeOverride(MethodBinding currentMethod, MethodBinding inheritedMethod) {
    if (inheritedMethod.declaringClass.isRawType())
      return true; 
    MethodBinding originalInherited = inheritedMethod.original();
    TypeBinding originalInheritedReturnType = originalInherited.returnType.leafComponentType();
    if (originalInheritedReturnType.isParameterizedTypeWithActualArguments())
      return !currentMethod.returnType.leafComponentType().isRawType(); 
    TypeBinding currentReturnType = currentMethod.returnType.leafComponentType();
    switch (currentReturnType.kind()) {
      case 4100:
        if (TypeBinding.equalsEquals(currentReturnType, inheritedMethod.returnType.leafComponentType()))
          return true; 
        break;
    } 
    if (originalInheritedReturnType.isTypeVariable() && 
      ((TypeVariableBinding)originalInheritedReturnType).declaringElement == originalInherited)
      return false; 
    return true;
  }
  
  boolean isInterfaceMethodImplemented(MethodBinding inheritedMethod, MethodBinding existingMethod, ReferenceBinding superType) {
    if (inheritedMethod.original() != inheritedMethod && existingMethod.declaringClass.isInterface())
      return false; 
    inheritedMethod = computeSubstituteMethod(inheritedMethod, existingMethod);
    if (inheritedMethod == null || !doesMethodOverride(existingMethod, inheritedMethod))
      return false; 
    return !(!TypeBinding.equalsEquals(inheritedMethod.returnType, existingMethod.returnType) && (
      !TypeBinding.notEquals((TypeBinding)this.type, (TypeBinding)existingMethod.declaringClass) || 
      existingMethod.declaringClass.isInterface() || 
      !areReturnTypesCompatible(existingMethod, inheritedMethod)));
  }
  
  public boolean isMethodSubsignature(MethodBinding method, MethodBinding inheritedMethod) {
    if (!CharOperation.equals(method.selector, inheritedMethod.selector))
      return false; 
    if (method.declaringClass.isParameterizedType())
      method = method.original(); 
    MethodBinding inheritedOriginal = method.findOriginalInheritedMethod(inheritedMethod);
    return isParameterSubsignature(method, (inheritedOriginal == null) ? inheritedMethod : inheritedOriginal);
  }
  
  boolean isUnsafeReturnTypeOverride(MethodBinding currentMethod, MethodBinding inheritedMethod) {
    if (TypeBinding.equalsEquals(currentMethod.returnType, inheritedMethod.returnType.erasure())) {
      TypeBinding[] currentParams = currentMethod.parameters;
      TypeBinding[] inheritedParams = inheritedMethod.parameters;
      for (int i = 0, l = currentParams.length; i < l; i++) {
        if (!areTypesEqual(currentParams[i], inheritedParams[i]))
          return true; 
      } 
    } 
    if (currentMethod.typeVariables == Binding.NO_TYPE_VARIABLES && 
      (inheritedMethod.original()).typeVariables != Binding.NO_TYPE_VARIABLES && 
      currentMethod.returnType.erasure().findSuperTypeOriginatingFrom(inheritedMethod.returnType.erasure()) != null)
      return true; 
    return false;
  }
  
  boolean reportIncompatibleReturnTypeError(MethodBinding currentMethod, MethodBinding inheritedMethod) {
    if (isUnsafeReturnTypeOverride(currentMethod, inheritedMethod)) {
      problemReporter(currentMethod).unsafeReturnTypeOverride(currentMethod, inheritedMethod, this.type);
      return false;
    } 
    return super.reportIncompatibleReturnTypeError(currentMethod, inheritedMethod);
  }
  
  void verify() {
    if (this.type.isAnnotationType())
      this.type.detectAnnotationCycle(); 
    super.verify();
    reportRawReferences();
    for (int i = this.type.typeVariables.length; --i >= 0; ) {
      TypeVariableBinding var = this.type.typeVariables[i];
      if (var.superInterfaces == Binding.NO_SUPERINTERFACES || (
        var.superInterfaces.length == 1 && var.superclass.id == 1))
        continue; 
      this.currentMethods = new HashtableOfObject(0);
      ReferenceBinding superclass = var.superclass();
      if (superclass.kind() == 4100)
        superclass = (ReferenceBinding)superclass.erasure(); 
      ReferenceBinding[] itsInterfaces = var.superInterfaces();
      ReferenceBinding[] superInterfaces = new ReferenceBinding[itsInterfaces.length];
      for (int j = itsInterfaces.length; --j >= 0;)
        superInterfaces[j] = (itsInterfaces[j].kind() == 4100) ? 
          (ReferenceBinding)itsInterfaces[j].erasure() : 
          itsInterfaces[j]; 
      computeInheritedMethods(superclass, superInterfaces);
      checkTypeVariableMethods(this.type.scope.referenceContext.typeParameters[i]);
    } 
  }
}
